using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;
using Microsoft.Xna.Framework.Graphics;
using TiledLib;

namespace RuminateContentPipeline {

    // For the map itself, we just store the size, tile size, and a list of layers.
    [ContentSerializerRuntimeType("Ruminate.Tiled.TiledMap, RuminateGameEngine")]
    public class DemoMapContent {
        public int Width;
        public int Height;
        public int TileWidth;
        public int TileHeight;
        public List<DemoMapTileContent[]> TileLayers;
        public List<Vector2[]> CollisionObjects;
    }

    // Each tile has a texture, source rect, and sprite effects.
    [ContentSerializerRuntimeType("Ruminate.Tiled.Tile, RuminateGameEngine")]
    public class DemoMapTileContent {
        public ExternalReference<Texture2DContent> Texture;
        public Rectangle SourceRectangle;
        public SpriteEffects SpriteEffects;
    }

    [ContentProcessor(DisplayName = "Tiled Importer - Ruminate")]
    public class MapProcessor : ContentProcessor<MapContent, DemoMapContent> {
        public override DemoMapContent Process(MapContent input, ContentProcessorContext context) {

            //System.Diagnostics.Debugger.Launch();

            // build the textures
            TiledHelpers.BuildTileSetTextures(input, context);

            // generate source rectangles
            TiledHelpers.GenerateTileSourceRectangles(input);

            // now build our output, first by just copying over some data
            var output = new DemoMapContent {
                Width = input.Layers[0].Width,
                Height = input.Layers[0].Height,
                TileWidth = input.TileWidth,
                TileHeight = input.TileHeight,
                TileLayers = new List<DemoMapTileContent[]>(),
                CollisionObjects = new List<Vector2[]>()
            };

            // iterate all the layers of the input
            foreach (LayerContent layer in input.Layers) {

                var olc = layer as MapObjectLayerContent;
                if (olc != null) {                    
                    foreach (var obj in olc.Objects) {                        
                        if (obj.ObjectType == MapObjectType.Polygon) {
                            for (var x = 0; x < obj.Points.Count; x++ ) {
                                obj.Points[x] = new Vector2(
                                    obj.Points[x].X + obj.Bounds.X,
                                    obj.Points[x].Y + obj.Bounds.Y);
                            }
                            output.CollisionObjects.Add(obj.Points.ToArray());
                        }
                    }
                }

                // we only care about tile layers in our demo
                var tlc = layer as TileLayerContent;
                if (tlc != null) {                    

                    // create the new layer
                    var layerTiles = new DemoMapTileContent[tlc.Data.Length];

                    // we need to build up our tile list now
                    for (int i = 0; i < tlc.Data.Length; i++) {
                        // get the ID of the tile
                        var tileId = tlc.Data[i];

                        // use that to get the actual index as well as the SpriteEffects
                        int tileIndex;
                        SpriteEffects spriteEffects;
                        TiledHelpers.DecodeTileID(tileId, out tileIndex, out spriteEffects);

                        //Draw these tiles
                        if (tileIndex > 0) {
                            // figure out which tile set has this tile index in it and grab
                            // the texture reference and source rectangle.
                            ExternalReference<Texture2DContent> textureContent = null;
                            var sourceRect = new Rectangle();

                            // iterate all the tile sets
                            foreach (var tileSet in input.TileSets) {
                                // if our tile index is in this set
                                if (tileIndex - tileSet.FirstId >= tileSet.Tiles.Count) { continue; }

                                // store the texture content and source rectangle
                                textureContent = tileSet.Texture;
                                sourceRect = tileSet.Tiles[(tileIndex - tileSet.FirstId)].Source;

                                // and break out of the foreach loop
                                break;
                            }

                            // now insert the tile into our output
                            layerTiles[i] = new DemoMapTileContent {
                                Texture = textureContent,
                                SourceRectangle = sourceRect,
                                SpriteEffects = spriteEffects
                            };
                        } else {
                            //Don't draw this tile
                            layerTiles[i] = null;
                        }                        
                    }

                    // add the layer to our output
                    output.TileLayers.Add(layerTiles);
                }
            }

            // return the output object. because we have ContentSerializerRuntimeType attributes on our
            // objects, we don't need a ContentTypeWriter and can just use the automatic serialization.
            return output;
        }
    }
}
